"""
Vendored from `pyethereum.abi`
"""
import warnings

from io import BytesIO

from eth_utils import (
    is_text,
    force_bytes,
    remove_0x_prefix,
    decode_hex,
)

from eth_abi.decoding import (
    get_single_decoder,
    get_multi_decoder,
)
from eth_abi.encoding import (
    get_single_encoder,
    get_multi_encoder,
)

from eth_abi.utils.parsing import (
    process_type,
)


def encode_single(typ, arg):
    try:
        base, sub, arrlist = typ
    except ValueError:
        base, sub, arrlist = process_type(typ)

    if is_text(arg):
        arg = force_bytes(arg)

    encoder = get_single_encoder(base, sub, arrlist)
    return encoder(arg)


def encode_abi(types, args):
    processed_types = [process_type(typ) for typ in types]
    encoder = get_multi_encoder(processed_types)
    return encoder(args)


HEX_CHARS = b'1234567890abcdef'


def is_hex_encoded_value(v):
    if not remove_0x_prefix(force_bytes(v)).lower().strip(HEX_CHARS) == b'':
        return False
    if len(remove_0x_prefix(v)) % 64 and len(remove_0x_prefix(v)) % 40:
        return False
    return True


# Decodes a single base datum
def decode_single(typ, data):
    if is_hex_encoded_value(data):
        warnings.warn(DeprecationWarning(
            "Automatic inference of hex encoded data has been deprecated. "
            "Please adjust your code to ensure that the data argument for "
            "`decode_single` is a byte string"
        ))
        data = decode_hex(remove_0x_prefix(data))

    if is_text(data):
        warnings.warn(DeprecationWarning(
            "Automatic conversion of encoded data to bytes has been deprecated. "
            "Please adjust your code to ensure that the data argument for "
            "`decode_single` is a byte string"
        ))
        data = force_bytes(data)

    try:
        base, sub, arrlist = typ
    except ValueError:
        base, sub, arrlist = process_type(typ)

    decoder = get_single_decoder(base, sub, arrlist)
    stream = BytesIO(data)
    return decoder(stream)


# Decodes multiple arguments using the head/tail mechanism
def decode_abi(types, data):
    if is_hex_encoded_value(data):
        warnings.warn(DeprecationWarning(
            "Automatic inference of hex encoded data has been deprecated. "
            "Please adjust your code to ensure that the data argument for "
            "`decode_single` is a byte string"
        ))
        data = decode_hex(remove_0x_prefix(data))

    if is_text(data):
        warnings.warn(DeprecationWarning(
            "Automatic conversion of encoded data to bytes has been deprecated. "
            "Please adjust your code to ensure that the data argument for "
            "`decode_abi` is a byte string"
        ))
        data = force_bytes(data)

    processed_types = tuple(process_type(_type) for _type in types)
    decoder = get_multi_decoder(processed_types)
    stream = BytesIO(data)
    return decoder(stream)



import itertools

from eth_utils import (
    is_boolean,
    is_integer,
    is_number,
    is_address,
    is_bytes,
    is_list_like,
    is_null,
    to_canonical_address,
)

from eth_abi.exceptions import (
    EncodingTypeError,
    ValueOutOfBounds,
)

from eth_abi.utils.numeric import (
    int_to_big_endian,
    compute_signed_integer_bounds,
    compute_unsigned_integer_bounds,
    compute_signed_real_bounds,
    compute_unsigned_real_bounds,
    ceil32,
)
from eth_abi.utils.padding import (
    fpad,
    zpad,
    zpad_right,
)


def get_multi_encoder(processed_types):
    """
    """
    encoders = tuple(
        get_single_encoder(base, sub, arrlist) for base, sub, arrlist in processed_types
    )
    return MultiEncoder.as_encoder(encoders=encoders)


def get_single_encoder(base, sub, arrlist):
    if arrlist:
        item_encoder = get_single_encoder(base, sub, arrlist[:-1])
        if arrlist[-1]:
            return SizedArrayEncoder.as_encoder(
                array_size=arrlist[-1][0],
                item_encoder=item_encoder,
            )
        else:
            return DynamicArrayEncoder.as_encoder(item_encoder=item_encoder)
    elif base == 'address':
        return encode_address
    elif base == 'bool':
        return encode_bool
    elif base == 'bytes':
        if sub:
            return BytesEncoder.as_encoder(value_bit_size=int(sub) * 8)
        else:
            return encode_bytes
    elif base == 'int':
        return SignedIntegerEncoder.as_encoder(value_bit_size=int(sub))
    elif base == 'string':
        return encode_string
    elif base == 'uint':
        return UnsignedIntegerEncoder.as_encoder(value_bit_size=int(sub))
    elif base == 'ureal':
        high_bit_size, low_bit_size = [int(v) for v in sub.split('x')]
        return UnsignedRealEncoder.as_encoder(
            value_bit_size=high_bit_size + low_bit_size,
            high_bit_size=high_bit_size,
            low_bit_size=low_bit_size,
        )
    elif base == 'real':
        high_bit_size, low_bit_size = [int(v) for v in sub.split('x')]
        return SignedRealEncoder.as_encoder(
            value_bit_size=high_bit_size + low_bit_size,
            high_bit_size=high_bit_size,
            low_bit_size=low_bit_size,
        )
    else:
        raise ValueError(
            "Unsupported type: {0} - must be one of "
            "address/bool/bytesXX/bytes/string/uintXXX/intXXX"
        )


class BaseEncoder(object):
    @classmethod
    def as_encoder(cls, name=None, **kwargs):
        for key in kwargs:
            if not hasattr(cls, key):
                raise AttributeError(
                    "Property {0} not found on Decoder class. "
                    "`Decoder.factory` only accepts keyword arguments which are "
                    "present on the Decoder class".format(key)
                )
        if name is None:
            name = cls.__name__
        sub_cls = type(name, (cls,), kwargs)
        sub_cls.validate()
        instance = sub_cls()
        return instance

    @classmethod
    def validate(cls):
        pass

    def __call__(self, value):
        return self.encode(value)


class MultiEncoder(BaseEncoder):
    encoders = None

    @classmethod
    def validate(cls):
        super(MultiEncoder, cls).validate()
        if cls.encoders is None:
            raise ValueError("`encoders` may not be none")

    @classmethod
    def encode(cls, values):
        if len(values) != len(cls.encoders):
            raise ValueOutOfBounds(
                "Recieved {0} values to encode.  Expected {1}".format(
                    len(values),
                    len(cls.encoders),
                )
            )
        raw_head_chunks = []
        tail_chunks = []

        for value, encoder in zip(values, cls.encoders):
            if isinstance(encoder, (DynamicArrayEncoder, StringEncoder)):
                raw_head_chunks.append(None)
                tail_chunks.append(encoder(value))
            else:
                raw_head_chunks.append(encoder(value))
                tail_chunks.append(b'')

        head_length = sum((
            32 if is_null(item) else len(item)
            for item in raw_head_chunks
        ))
        tail_offsets = tuple((
            sum((len(chunk) for chunk in tail_chunks[:i]))
            for i in range(len(tail_chunks))
        ))
        head_chunks = tuple((
            (
                encode_uint_256(head_length + tail_offsets[idx])
                if is_null(head_chunk)
                else head_chunk
            ) for idx, head_chunk
            in enumerate(raw_head_chunks)
        ))
        encoded_value = b''.join(tuple(itertools.chain(head_chunks, tail_chunks)))
        return encoded_value


class FixedSizeEncoder(BaseEncoder):
    value_bit_size = None
    data_byte_size = None
    encode_fn = None
    type_check_fn = None
    is_big_endian = None

    @classmethod
    def validate(cls):
        super(FixedSizeEncoder, cls).validate()
        if cls.value_bit_size is None:
            raise ValueError("`value_bit_size` may not be none")
        if cls.data_byte_size is None:
            raise ValueError("`data_byte_size` may not be none")
        if cls.encode_fn is None:
            raise ValueError("`encode_fn` may not be none")
        if cls.is_big_endian is None:
            raise ValueError("`is_big_endian` may not be none")

        if cls.value_bit_size % 8 != 0:
            raise ValueError(
                "Invalid value bit size: {0}.  Must be a multiple of 8".format(
                    cls.value_bit_size,
                )
            )

        if cls.value_bit_size > cls.data_byte_size * 8:
            raise ValueError("Value byte size exceeds data size")

    @classmethod
    def validate_value(cls, value):
        raise NotImplementedError("Must be implemented by subclasses")

    @classmethod
    def encode(cls, value):
        cls.validate_value(value)
        base_encoded_value = cls.encode_fn(value)

        if cls.is_big_endian:
            padded_encoded_value = zpad(base_encoded_value, cls.data_byte_size)
        else:
            padded_encoded_value = zpad_right(base_encoded_value, cls.data_byte_size)
        return padded_encoded_value


class Fixed32ByteSizeEncoder(FixedSizeEncoder):
    data_byte_size = 32


class BooleanEncoder(Fixed32ByteSizeEncoder):
    value_bit_size = 8
    is_big_endian = True

    @classmethod
    def validate_value(cls, value):
        if not is_boolean(value):
            raise EncodingTypeError(
                "Value of type {0} cannot be encoded by {0}".format(
                    type(value),
                    cls.__name__,
                )
            )

    @classmethod
    def encode_fn(cls, value):
        if value is True:
            return b'\x01'
        elif value is False:
            return b'\x00'
        else:
            raise ValueError("Invariant")


encode_bool = BooleanEncoder.as_encoder()


class NumberEncoder(Fixed32ByteSizeEncoder):
    is_big_endian = True
    bounds_fn = None
    type_check_fn = None

    @classmethod
    def validate(cls):
        super(NumberEncoder, cls).validate()
        if cls.bounds_fn is None:
            raise ValueError("`bounds_fn` cannot be null")
        if cls.type_check_fn is None:
            raise ValueError("`type_check_fn` cannot be null")

    @classmethod
    def validate_value(cls, value):
        if not cls.type_check_fn(value):
            raise EncodingTypeError(
                "Value of type {0} cannot be encoded by {0}".format(
                    type(value),
                    cls.__name__,
                )
            )

        lower_bound, upper_bound = cls.bounds_fn(cls.value_bit_size)

        if value < lower_bound or value > upper_bound:
            raise ValueOutOfBounds(
                "Value '{0}' cannot be encoded in {1} bits.  Must be bounded "
                "between [{2}, {3}]".format(
                    value,
                    cls.value_bit_size,
                    lower_bound,
                    upper_bound,
                )
            )


class UnsignedIntegerEncoder(NumberEncoder):
    encode_fn = staticmethod(int_to_big_endian)
    bounds_fn = staticmethod(compute_unsigned_integer_bounds)
    type_check_fn = staticmethod(is_integer)


encode_uint_256 = UnsignedIntegerEncoder.as_encoder(value_bit_size=256, data_byte_size=32)


class SignedIntegerEncoder(NumberEncoder):
    bounds_fn = staticmethod(compute_signed_integer_bounds)
    type_check_fn = staticmethod(is_integer)

    @classmethod
    def encode_fn(cls, value):
        return int_to_big_endian(value % 2**cls.value_bit_size)

    @classmethod
    def encode(cls, value):
        cls.validate_value(value)
        base_encoded_value = cls.encode_fn(value)

        if value >= 0:
            padded_encoded_value = zpad(base_encoded_value, cls.data_byte_size)
        else:
            padded_encoded_value = fpad(base_encoded_value, cls.data_byte_size)
        return padded_encoded_value


class BaseRealEncoder(NumberEncoder):
    low_bit_size = None
    high_bit_size = None
    type_check_fn = staticmethod(is_number)

    @classmethod
    def validate(cls):
        super(BaseRealEncoder, cls).validate()
        if cls.high_bit_size is None:
            raise ValueError("`high_bit_size` cannot be null")
        if cls.low_bit_size is None:
            raise ValueError("`low_bit_size` cannot be null")
        if cls.low_bit_size + cls.high_bit_size != cls.value_bit_size:
            raise ValueError("high and low bitsizes must sum to the value_bit_size")


class UnsignedRealEncoder(BaseRealEncoder):
    @classmethod
    def bounds_fn(cls, value_bit_size):
        return compute_unsigned_real_bounds(cls.high_bit_size, cls.low_bit_size)

    @classmethod
    def encode_fn(cls, value):
        scaled_value = value * 2 ** cls.low_bit_size
        integer_value = int(scaled_value)
        return int_to_big_endian(integer_value)


class SignedRealEncoder(BaseRealEncoder):
    @classmethod
    def bounds_fn(cls, value_bit_size):
        return compute_signed_real_bounds(cls.high_bit_size, cls.low_bit_size)

    @classmethod
    def encode_fn(cls, value):
        scaled_value = value * 2 ** cls.low_bit_size
        integer_value = int(scaled_value)
        unsigned_integer_value = integer_value % 2 ** (cls.high_bit_size + cls.low_bit_size)
        return int_to_big_endian(unsigned_integer_value)

    @classmethod
    def encode(cls, value):
        cls.validate_value(value)
        base_encoded_value = cls.encode_fn(value)

        if value >= 0:
            padded_encoded_value = zpad(base_encoded_value, cls.data_byte_size)
        else:
            padded_encoded_value = fpad(base_encoded_value, cls.data_byte_size)
        return padded_encoded_value


class AddressEncoder(Fixed32ByteSizeEncoder):
    value_bit_size = 20 * 8
    encode_fn = staticmethod(to_canonical_address)
    is_big_endian = True

    @classmethod
    def validate_value(cls, value):
        if not is_address(value):
            raise EncodingTypeError(
                "Value of type {0} cannot be encoded by {0}".format(
                    type(value),
                    cls.__name__,
                )
            )

    @classmethod
    def validate(cls):
        super(AddressEncoder, cls).validate()
        if cls.value_bit_size != 20 * 8:
            raise ValueError('Addresses must be 160 bits in length')


encode_address = AddressEncoder.as_encoder()


class BytesEncoder(Fixed32ByteSizeEncoder):
    is_big_endian = False

    @classmethod
    def validate_value(cls, value):
        if not is_bytes(value):
            raise EncodingTypeError(
                "Value of type {0} cannot be encoded by {0}".format(
                    type(value),
                    cls.__name__,
                )
            )
        if len(value) > cls.value_bit_size // 8:
            raise ValueOutOfBounds(
                "String {0} exceeds total byte size for bytes{1} encoding".format(
                    value,
                    cls.value_bit_size // 8,
                )
            )

    @classmethod
    def encode_fn(cls, value):
        return value


class StringEncoder(BaseEncoder):
    @classmethod
    def encode(cls, value):
        if not is_bytes(value):
            raise EncodingTypeError(
                "Value of type {0} cannot be encoded as a string".format(
                    type(value),
                )
            )

        encoded_size = encode_uint_256(len(value))
        if not value:
            padded_value = b'\x00' * 32
        else:
            padded_value = zpad_right(value, ceil32(len(value)))
        encoded_value = encoded_size + padded_value

        return encoded_value


encode_string = encode_bytes = StringEncoder.as_encoder()


class BaseArrayEncoder(BaseEncoder):
    item_encoder = None

    @classmethod
    def validate(cls):
        super(BaseArrayEncoder, cls).validate()
        if cls.item_encoder is None:
            raise ValueError("`item_encoder` may not be none")

    @classmethod
    def encode_elements(cls, value):
        if not is_list_like(value):
            raise EncodingTypeError(
                "Cannot encode value of type {0} using array encoder.  Must be "
                "a list-like object such as an array or tuple".format(
                    type(value),
                )
            )
        encoded_elements = b''.join((
            cls.item_encoder(item)
            for item in value
        ))
        return encoded_elements


class SizedArrayEncoder(BaseArrayEncoder):
    array_size = None

    @classmethod
    def validate(cls):
        super(SizedArrayEncoder, cls).validate()
        if cls.array_size is None:
            raise ValueError("`array_size` may not be none")

    @classmethod
    def encode(cls, value):
        if len(value) != cls.array_size:
            raise ValueOutOfBounds(
                "Expected value with length {0}.  Provided value has {1} "
                "elements".format(cls.array_size, len(value))
            )
        encoded_elements = cls.encode_elements(value)
        return encoded_elements


class DynamicArrayEncoder(BaseArrayEncoder):
    @classmethod
    def encode(cls, value):
        encoded_size = encode_uint_256(len(value))
        encoded_elements = cls.encode_elements(value)
        encoded_value = encoded_size + encoded_elements
        return encoded_value



import re
import ast

from eth_utils import (
    force_text,
)


def process_type(typ):
    # Crazy reg expression to separate out base type component (eg. uint),
    # size (eg. 256, 128x128, none), array component (eg. [], [45], none)
    regexp = '([a-z]*)([0-9]*x?[0-9]*)((\[[0-9]*\])*)'
    base, sub, arr, _ = re.match(regexp, force_text(typ)).groups()
    arrlist = re.findall('\[[0-9]*\]', arr)
    if len(''.join(arrlist)) != len(arr):
        raise ValueError("Unknown characters found in array declaration")
    # Check validity of string type
    if base == 'string' or base == 'bytes':
        if not re.match('^[0-9]*$', sub):
            raise ValueError("String type must have no suffix or numerical suffix")
        if sub and int(sub) > 32:
            raise ValueError("Maximum 32 bytes for fixed-length str or bytes")
    # Check validity of integer type
    elif base == 'uint' or base == 'int':
        if not re.match('^[0-9]+$', sub):
            raise ValueError("Integer type must have numerical suffix")
        if 8 > int(sub) or int(sub) > 256:
            raise ValueError("Integer size out of bounds")
        if int(sub) % 8 != 0:
            raise ValueError("Integer size must be multiple of 8")
    # Check validity of real type
    elif base == 'ureal' or base == 'real':
        if not re.match('^[0-9]+x[0-9]+$', sub):
            raise ValueError("Real type must have suffix of form <high>x<low>, eg. 128x128")
        high, low = [int(x) for x in sub.split('x')]
        if 8 > (high + low) or (high + low) > 256:
            raise ValueError("Real size out of bounds (max 32 bytes)")
        if high % 8 != 0 or low % 8 != 0:
            raise ValueError("Real high/low sizes must be multiples of 8")
    # Check validity of hash type
    elif base == 'hash':
        if not re.match('^[0-9]+$', sub):
            raise ValueError("Hash type must have numerical suffix")
    # Check validity of address type
    elif base == 'address':
        if sub != '':
            raise ValueError("Address cannot have suffix")
    return base, sub, [ast.literal_eval(x) for x in arrlist]



import sys
import math
import decimal

from eth_utils import (
    encode_hex,
    is_number,
)

from eth_abi.constants import (
    TT256,
)


if sys.version_info.major == 2:
    import struct

    def int_to_big_endian(value):
        cs = []
        while value > 0:
            cs.append(chr(value % 256))
            value /= 256
        s = ''.join(reversed(cs))
        return s

    def big_endian_to_int(value):
        if len(value) == 1:
            return ord(value)
        elif len(value) <= 8:
            return struct.unpack('>Q', value.rjust(8, '\x00'))[0]
        else:
            return int(encode_hex(value), 16)
else:
    def int_to_big_endian(value):
        byte_length = math.ceil(value.bit_length() / 8)
        return (value).to_bytes(byte_length, byteorder='big')

    def big_endian_to_int(value):
        return int.from_bytes(value, byteorder='big')


def ceil32(x):
    return x if x % 32 == 0 else x + 32 - (x % 32)


def encode_int(value):
    '''encodes an integer into serialization'''
    if not is_number(value) or value < 0 or value >= TT256:
        raise Exception("Integer invalid or out of range: %r" % value)
    return int_to_big_endian(value)


def compute_unsigned_integer_bounds(num_bits):
    return (
        0,
        2 ** num_bits - 1,
    )


def compute_signed_integer_bounds(num_bits):
    return (
        -1 * 2 ** (num_bits - 1),
        2 ** (num_bits - 1) - 1,
    )


def compute_unsigned_real_bounds(num_high_bits, num_low_bits):
    integer_lower_bound, integer_upper_bount = compute_unsigned_integer_bounds(
        num_high_bits,
    )
    return (
        integer_lower_bound * 1.0 / 2 ** num_low_bits,
        integer_upper_bount * 1.0 / 2 ** num_low_bits,
    )


def compute_signed_real_bounds(num_high_bits, num_low_bits):
    integer_lower_bound, integer_upper_bount = compute_signed_integer_bounds(
        num_high_bits,
    )
    return (
        integer_lower_bound * 1.0 / 2 ** num_low_bits,
        integer_upper_bount * 1.0 / 2 ** num_low_bits,
    )


def quantize_value(value, decimal_bit_size):
        num_decimals = int(math.ceil(math.log10(2 ** decimal_bit_size)))
        if num_decimals == 0:
            quantize_value = decimal.Decimal('1')
        else:
            quantize_value = decimal.Decimal('1.{0}'.format(''.zfill(num_decimals)))
        decimal_value = decimal.Decimal(value)
        return decimal_value.quantize(quantize_value)



class EncodingError(Exception):
    pass


class EncodingTypeError(EncodingError):
    """
    Raised when trying to encode a value which is of the wrong type for the
    desired encoding type.
    """
    pass


class ValueOutOfBounds(EncodingError):
    """
    Raised when trying to encode a value which is out bounds for the desired
    type.
    """
    pass


class DecodingError(Exception):
    pass


class InsufficientDataBytes(DecodingError):
    pass


class NonEmptyPaddingBytes(DecodingError):
    pass



import functools

from eth_utils import (
    pad_left,
    pad_right,
)


zpad = functools.partial(pad_left, pad_with='\x00')
zpad32 = functools.partial(pad_left, to_size=32, pad_with='\x00')
zpad_right = functools.partial(pad_right, pad_with='\x00')
zpad32_right = functools.partial(pad_right, to_size=32, pad_with='\x00')

fpad = functools.partial(pad_left, pad_with='\xff')
fpad32 = functools.partial(pad_left, to_size=32, pad_with='\xff')



from __future__ import unicode_literals

import decimal

from eth_utils import (
    force_text,
    to_tuple,
    to_normalized_address,
)

from eth_abi.exceptions import (
    InsufficientDataBytes,
    NonEmptyPaddingBytes,
)
from eth_abi.utils.numeric import (
    big_endian_to_int,
    quantize_value,
    ceil32,
)


decimal.DefaultContext.prec = 999


def get_multi_decoder(processed_types):
    """
    """
    decoders = tuple(
        get_single_decoder(base, sub, arrlist) for base, sub, arrlist in processed_types
    )
    return MultiDecoder.as_decoder(decoders=decoders)


def get_single_decoder(base, sub, arrlist):
    if arrlist:
        item_decoder = get_single_decoder(base, sub, arrlist[:-1])
        if arrlist[-1]:
            return SizedArrayDecoder.as_decoder(
                array_size=arrlist[-1][0],
                item_decoder=item_decoder,
            )
        else:
            return DynamicArrayDecoder.as_decoder(item_decoder=item_decoder)
    elif base == 'address':
        return decode_address
    elif base == 'bool':
        return decode_bool
    elif base == 'bytes':
        if sub:
            return BytesDecoder.as_decoder(value_bit_size=int(sub) * 8)
        else:
            return decode_bytes
    elif base == 'int':
        return SignedIntegerDecoder.as_decoder(value_bit_size=int(sub))
    elif base == 'string':
        return decode_string
    elif base == 'uint':
        return UnsignedIntegerDecoder.as_decoder(value_bit_size=int(sub))
    elif base == 'ureal':
        high_bit_size, low_bit_size = [int(v) for v in sub.split('x')]
        return UnsignedRealDecoder.as_decoder(
            value_bit_size=high_bit_size + low_bit_size,
            high_bit_size=high_bit_size,
            low_bit_size=low_bit_size,
        )
    elif base == 'real':
        high_bit_size, low_bit_size = [int(v) for v in sub.split('x')]
        return SignedRealDecoder.as_decoder(
            value_bit_size=high_bit_size + low_bit_size,
            high_bit_size=high_bit_size,
            low_bit_size=low_bit_size,
        )
    else:
        raise ValueError(
            "Unsupported type: {0} - must be one of "
            "address/bool/bytesXX/bytes/string/uintXXX/intXXX"
        )


class BaseDecoder(object):
    @classmethod
    def as_decoder(cls, name=None, **kwargs):
        for key in kwargs:
            if not hasattr(cls, key):
                raise AttributeError(
                    "Property {0} not found on Decoder class. "
                    "`Decoder.factory` only accepts keyword arguments which are "
                    "present on the Decoder class".format(key)
                )
        if name is None:
            name = cls.__name__
        sub_cls = type(name, (cls,), kwargs)
        sub_cls.validate()
        instance = sub_cls()
        return instance

    @classmethod
    def validate(cls):
        pass

    def __call__(self, stream):
        return self.decode(stream)


class HeadTailDecoder(BaseDecoder):
    tail_decoder = None

    @classmethod
    def validate(cls):
        super(HeadTailDecoder, cls).validate()
        if cls.tail_decoder is None:
            raise ValueError("No `tail_decoder` set")

    @classmethod
    def decode(cls, stream):
        start_pos = decode_uint_256(stream)
        anchor_pos = stream.tell()
        stream.seek(start_pos)
        value = cls.tail_decoder(stream)
        stream.seek(anchor_pos)
        return value


class MultiDecoder(BaseDecoder):
    decoders = None

    @classmethod
    def validate(cls):
        super(MultiDecoder, cls).validate()
        if cls.decoders is None:
            raise ValueError("No `decoders` set")

    @classmethod
    @to_tuple
    def decode(cls, stream):
        for decoder in cls.decoders:
            if isinstance(decoder, (DynamicArrayDecoder, StringDecoder)):
                yield HeadTailDecoder.as_decoder(tail_decoder=decoder)(stream)
            else:
                yield decoder(stream)


class SingleDecoder(BaseDecoder):
    decoder_fn = None

    @classmethod
    def validate(cls):
        super(SingleDecoder, cls).validate()
        if cls.decoder_fn is None:
            raise ValueError("No `decoder_fn` set")

    @classmethod
    def validate_padding_bytes(cls, value, padding_bytes):
        raise NotImplementedError("Must be implemented by subclasses")
        value_byte_size = cls._get_value_byte_size()
        padding_size = cls.data_byte_size - value_byte_size

        if padding_bytes != b'\x00' * padding_size:
            raise NonEmptyPaddingBytes(
                "Padding bytes were not empty: {0}".format(force_text(padding_bytes))
            )

    @classmethod
    def decode(cls, stream):
        raw_data = cls.read_data_from_stream(stream)
        data, padding_bytes = cls.split_data_and_padding(raw_data)
        value = cls.decoder_fn(data)
        cls.validate_padding_bytes(value, padding_bytes)

        return value

    @classmethod
    def read_data_from_stream(cls, stream):
        raise NotImplementedError("Must be implemented by subclasses")

    @classmethod
    def split_data_and_padding(cls, raw_data):
        return raw_data, b''


class BaseArrayDecoder(BaseDecoder):
    item_decoder = None

    @classmethod
    def validate(cls):
        super(BaseArrayDecoder, cls).validate()
        if cls.item_decoder is None:
            raise ValueError("No `item_decoder` set")


class SizedArrayDecoder(BaseArrayDecoder):
    array_size = None

    @classmethod
    @to_tuple
    def decode(cls, stream):
        for _ in range(cls.array_size):
            yield cls.item_decoder(stream)


class DynamicArrayDecoder(BaseArrayDecoder):
    @classmethod
    @to_tuple
    def decode(cls, stream):
        array_size = decode_uint_256(stream)
        for _ in range(array_size):
            yield cls.item_decoder(stream)


class FixedByteSizeDecoder(SingleDecoder):
    decoder_fn = None
    value_bit_size = None
    data_byte_size = None
    is_big_endian = None

    @classmethod
    def validate(cls):
        super(FixedByteSizeDecoder, cls).validate()

        if cls.value_bit_size is None:
            raise ValueError("`value_bit_size` may not be None")
        if cls.data_byte_size is None:
            raise ValueError("`data_byte_size` may not be None")
        if cls.decoder_fn is None:
            raise ValueError("`decoder_fn` may not be None")
        if cls.is_big_endian is None:
            raise ValueError("`is_big_endian` may not be None")

        if cls.value_bit_size % 8 != 0:
            raise ValueError(
                "Invalid value bit size: {0}.  Must be a multiple of 8".format(
                    cls.value_bit_size,
                )
            )

        if cls.value_bit_size > cls.data_byte_size * 8:
            raise ValueError("Value byte size exceeds data size")

    @classmethod
    def read_data_from_stream(cls, stream):
        data = stream.read(cls.data_byte_size)

        if len(data) != cls.data_byte_size:
            raise InsufficientDataBytes(
                "Tried to read {0} bytes.  Only got {1} bytes".format(
                    cls.data_byte_size,
                    len(data),
                )
            )

        return data

    @classmethod
    def split_data_and_padding(cls, raw_data):
        value_byte_size = cls._get_value_byte_size()
        padding_size = cls.data_byte_size - value_byte_size

        if cls.is_big_endian:
            padding_bytes = raw_data[:padding_size]
            data = raw_data[padding_size:]
        else:
            data = raw_data[:value_byte_size]
            padding_bytes = raw_data[value_byte_size:]

        return data, padding_bytes

    @classmethod
    def validate_padding_bytes(cls, value, padding_bytes):
        value_byte_size = cls._get_value_byte_size()
        padding_size = cls.data_byte_size - value_byte_size

        if padding_bytes != b'\x00' * padding_size:
            raise NonEmptyPaddingBytes(
                "Padding bytes were not empty: {0}".format(force_text(padding_bytes))
            )

    @classmethod
    def _get_value_byte_size(cls):
        value_byte_size = cls.value_bit_size // 8
        return value_byte_size


class Fixed32ByteSizeDecoder(FixedByteSizeDecoder):
    data_byte_size = 32


class BooleanDecoder(Fixed32ByteSizeDecoder):
    value_bit_size = 8
    is_big_endian = True

    @classmethod
    def decoder_fn(cls, data):
        if data == b'\x00':
            return False
        elif data == b'\x01':
            return True
        else:
            raise NonEmptyPaddingBytes(
                "Boolean must be either 0x0 or 0x1.  Got: {0}".format(force_text(data))
            )


decode_bool = BooleanDecoder.as_decoder()


class AddressDecoder(Fixed32ByteSizeDecoder):
    value_bit_size = 20 * 8
    is_big_endian = True
    decoder_fn = staticmethod(to_normalized_address)


decode_address = AddressDecoder.as_decoder()


#
# Unsigned Integer Decoders
#
class UnsignedIntegerDecoder(Fixed32ByteSizeDecoder):
    decoder_fn = staticmethod(big_endian_to_int)
    is_big_endian = True


decode_uint_256 = UnsignedIntegerDecoder.as_decoder(value_bit_size=256)


#
# Signed Integer Decoders
#
class SignedIntegerDecoder(Fixed32ByteSizeDecoder):
    is_big_endian = True

    @classmethod
    def decoder_fn(cls, data):
        value = big_endian_to_int(data)
        if value >= 2 ** (cls.value_bit_size - 1):
            return value - 2 ** cls.value_bit_size
        else:
            return value

    @classmethod
    def validate_padding_bytes(cls, value, padding_bytes):
        value_byte_size = cls._get_value_byte_size()
        padding_size = cls.data_byte_size - value_byte_size

        if value >= 0:
            expected_padding_bytes = b'\x00' * padding_size
        else:
            expected_padding_bytes = b'\xff' * padding_size

        if padding_bytes != expected_padding_bytes:
            raise NonEmptyPaddingBytes(
                "Padding bytes were not empty: {0}".format(force_text(padding_bytes))
            )


#
# Bytes1..32
#
class BytesDecoder(Fixed32ByteSizeDecoder):
    is_big_endian = False

    @classmethod
    def decoder_fn(cls, data):
        return data


class BaseRealDecoder(Fixed32ByteSizeDecoder):
    high_bit_size = None
    low_bit_size = None
    data_byte_size = None
    is_big_endian = True

    @classmethod
    def validate(cls):
        super(BaseRealDecoder, cls).validate()

        if cls.high_bit_size is None:
            raise ValueError("`high_bit_size` cannot be null")
        if cls.low_bit_size is None:
            raise ValueError("`low_bit_size` cannot be null")
        if cls.low_bit_size + cls.high_bit_size != cls.value_bit_size:
            raise ValueError("high and low bitsizes must sum to the value_bit_size")


class UnsignedRealDecoder(BaseRealDecoder):
    @classmethod
    def decoder_fn(cls, data):
        value = big_endian_to_int(data)
        decimal_value = decimal.Decimal(value)
        raw_real_value = decimal_value / 2 ** cls.low_bit_size
        real_value = quantize_value(raw_real_value, cls.low_bit_size)
        return real_value


class SignedRealDecoder(BaseRealDecoder):
    @classmethod
    def decoder_fn(cls, data):
        value = big_endian_to_int(data)
        if value >= 2 ** (cls.high_bit_size + cls.low_bit_size - 1):
            signed_value = value - 2 ** (cls.high_bit_size + cls.low_bit_size)
        else:
            signed_value = value
        signed_decimal_value = decimal.Decimal(signed_value)
        raw_real_value = signed_decimal_value / 2 ** cls.low_bit_size
        real_value = quantize_value(raw_real_value, cls.low_bit_size)
        return real_value

    @classmethod
    def validate_padding_bytes(cls, value, padding_bytes):
        value_byte_size = cls._get_value_byte_size()
        padding_size = cls.data_byte_size - value_byte_size

        if value >= 0:
            expected_padding_bytes = b'\x00' * padding_size
        else:
            expected_padding_bytes = b'\xff' * padding_size

        if padding_bytes != expected_padding_bytes:
            raise NonEmptyPaddingBytes(
                "Padding bytes were not empty: {0}".format(force_text(padding_bytes))
            )


#
# String and Bytes
#
class StringDecoder(SingleDecoder):
    @classmethod
    def decoder_fn(cls, data):
        return data

    @classmethod
    def read_data_from_stream(cls, stream):
        data_length = decode_uint_256(stream)
        padded_length = ceil32(data_length)

        data = stream.read(padded_length)

        if len(data) < padded_length:
            raise InsufficientDataBytes(
                "Tried to read {0} bytes.  Only got {1} bytes".format(
                    padded_length,
                    len(data),
                )
            )

        padding_bytes = data[data_length:]

        if padding_bytes != b'\x00' * (padded_length - data_length):
            raise NonEmptyPaddingBytes(
                "Padding bytes were not empty: {0}".format(force_text(padding_bytes))
            )

        return data[:data_length]

    @classmethod
    def validate_padding_bytes(cls, value, padding_bytes):
        pass


decode_string = decode_bytes = StringDecoder.as_decoder()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os

from setuptools import (
    setup,
    find_packages,
)


DIR = os.path.dirname(os.path.abspath(__file__))

version = '0.4.0'

readme = open(os.path.join(DIR, 'README.md')).read()


setup(
    name='ethereum-abi-utils',
    version=version,
    description="""Ethereum ABI Utils""",
    long_description=readme,
    author='Piper Merriam',
    author_email='pipermerriam@gmail.com',
    url='https://github.com/pipermerriam/ethereum-abi-utils',
    include_package_data=True,
    install_requires=[
        'ethereum-utils>=0.2.0',
    ],
    py_modules=['eth_abi'],
    license="MIT",
    zip_safe=False,
    keywords='ethereum',
    packages=find_packages(exclude=["tests", "tests.*"]),
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
    ],
)
